from enum import Enum
from typing import List, Optional, TypeVar, Dict

from escape.core.objects import par, parameter_obj, functor_obj
import escape.scattering.material as material

from periodictable.util import require_keywords
import periodictable.xsf as ptx
import periodictable.nsf as ptn
from periodictable.constants import avogadro_number, electron_radius
from periodictable import formulas
from periodictable.nsf import neutron_wavelength, ABSORPTION_WAVELENGTH

__all__ = ["Source", "amorphous"]

T = TypeVar('T', parameter_obj, int, float)

class Source(Enum):
    XRAYS = 0,
    NEUTRONS = 1


@require_keywords
def _xray_sld(compound: str, density: Optional[T]=None,  scale: float = 1e-6,
              wavelength:Optional[float] = None, energy:Optional[float] = None):

    if isinstance(density, parameter_obj):
        compound = formulas.formula(compound, density=density.value)
    elif isinstance(density, (float, int)):
        compound = formulas.formula(compound, density=density)
    density = compound.density
    if wavelength is not None:
        energy = ptx.xray_energy(wavelength)
    if energy is None:
        raise ValueError("scattering calculation needs energy or wavelength")

    mass, sum_f1, sum_f2 = 0, 0, 0
    for element, quantity in compound.atoms.items():
        mass += element.mass*quantity
        f1, f2 = element.xray.scattering_factors(energy=energy)
        #print element, f1, f2, wavelength
        sum_f1 += f1*quantity
        sum_f2 += f2*quantity
    if mass == 0: # because the formula is empty
        return (parameter_obj.convert(0), parameter_obj.convert(0))

    coeff = 1e-8*scale*avogadro_number/mass
    rho = density*(coeff*sum_f1*electron_radius)
    irho = density*(-coeff*sum_f2*electron_radius)
    return rho, irho


@require_keywords
def _neutron_sld(compound: str,
                 density: Optional[T]=None,
                 scale: float = 1e-6):
    
    if isinstance(density, parameter_obj):
        compound = formulas.formula(compound, density=density.value)
    else:
        compound = formulas.formula(compound, density=density)
    if density is None:
        density = compound.density
    # Sum over the quantities
    molar_mass = num_atoms = 0
    sigma_a_div_wl = b_c = 0
    for element, quantity in compound.atoms.items():
        if not element.neutron.has_sld():
            return None, None, None
        #print element,quantity,element.neutron.b_c,element.neutron.absorption,element.neutron.total
        molar_mass += element.mass*quantity
        num_atoms += quantity
        sigma_a_div_wl += quantity * element.neutron.absorption
        b_c += quantity * element.neutron.b_c

    # If nothing to sum, return values for a vacuum.  This might be because
    # the material has no atoms or it might be because the density is zero.
    if molar_mass == 0:
        return (parameter_obj.convert(0), parameter_obj.convert(0))

    # Turn sums into scattering factors
    b_c /= num_atoms
    sigma_a_div_wl *= 1.0/ABSORPTION_WAVELENGTH/num_atoms

    # Compute number density
    #cell_volume = (molar_mass/compound.density)/avogadro_number*1e24 # (10^8 A/cm)^3
    number_density = num_atoms / molar_mass*avogadro_number*1e-24*scale #without mass density

    # Compute SLD
    sld_re = density*(number_density * b_c * 10)
    sld_im = density*(-number_density * sigma_a_div_wl / 2 * 0.01)
    
    return (sld_re, sld_im)


def amorphous(compound: str,
              source: Source,
              density: Optional[T]=None,
              sldm: Optional[T]=None,
              sldscale: float=1e-6,
              sldunit: str = "1/A^2",
              wavelength: Optional[float]=None,
              energy: Optional[float]=None,
              density_based: bool = False
              ):
    """
    Returns an instance of material_obj with scattering length density parameters based on chemical formula and density of material.
    For the calculation of initial values of SLD parameters the 'periodictable' package is used.
    
    .. versionchanged:: 0.9.1
    Arguments changed.
    
    :Parameters:
        `compound`: str
            Chemical formula or element name compatible with 'periodictable' package
        `source`: Source
            Source.XRAYS or Source.NEUTRONS
        `density`: parameter_obj, double, None
            Mass density of the material. If density has 'parameter_obj' type the scattering length densities 'sldr' and 'sldi'
            of the material object are parameters dependent on density parameter.
            If density is a float value and density_based is
            False the material SLDs are independent parameters with initial values calculated from the provided mass density value.
            If density is a float value density_based is True a new density parameter is created and SLDs depend on it.
            If density is None, 'periodictable' value is used and behavior is the same as with float value (provided only for periodic table elements).
        `sldm`: parameter_obj or double
            Magnetic scattering length density. Relevant only for magnetic materials and polarized neutron scattering.
        `sldscale`: float
            Scale for SLD parameters.
        `sldunit`: str
            Units for SLD parameters.
        `wavelength`: float or None
            Radiation wavelength. Relevant for X-rays only.
        `energy`: float or None
            Radiation energy. Relevant for X-rays only.
        `density_based`: bool
            If True and density is a float value material SLDs parameters are dependent on newly created density parameter.
    :return:
        Material object 
        
    """
    sldre = sldim = None
    scale_ = 1 if density is None else sldscale
#    params = []
    if isinstance(density, (float, int)):
        scale_ = 1
        if density_based:
            density = par(f"Density {compound}", density, units="g/cm^3")
#            params.append(density)
    if source==Source.XRAYS:
        sldre, sldim = _xray_sld(compound, scale=scale_, density=density,
                                     wavelength=wavelength, energy=energy)
    elif source==Source.NEUTRONS:
        sldre, sldim = _neutron_sld(compound, scale=scale_, density=density)
        
    if isinstance(density, (float, int)) or density is None:
        sldre = par(f"SLDRe {compound}", sldre, scale=sldscale, units=sldunit)
        sldim = par(f"SLDIm {compound}", sldim, scale=sldscale, units=sldunit)
#        params.append(sldre)
#        params.append(sldim)
    mat = None
    if sldm is None:
        mat = material.amorphous(compound, sldre, sldim)
    else:
        if isinstance(sldm, (float, int)):
            sldm = par(f"SLDMag {compound}", sldm, scale=sldscale, units=sldunit)
            #params.append(sldm)
        mat = material.amorphous(compound, sldre, sldim, sldm)
#    return [mat] + params if params else mat
    return mat 

