
import ipywidgets.widgets as widgets
from ipywidgets import AppLayout, Layout
from threading import Lock

from escape.utils.widgets.parameter import ParametersPanel
from escape.utils.widgets.plot import ErrorPlot2dWidget, MapWidget, PLOT_TYPE_KW,\
    PLOT2D_TYPE, MAP_TYPE, PLOT_TYPES


def lock_update(func):
    def wrapper(self, *args, **kwargs):
        self.lock.acquire()
        try:
            return func(self, *args, **kwargs)
        finally:
            self.lock.release()
    return wrapper


class EscapeObjectLayout(AppLayout):
    
    center_widget_cls = None
    
    def __init__(self, escobj, parameter_callback=None, **kwargs):
        AppLayout.__init__(self, **kwargs)
        if parameter_callback is None:
            parameter_callback = self.update
        if not callable(parameter_callback):
            raise TypeError(
                ("Parameter callback is not callable,",
                 "use function or class instance with __call__ method"
                 ))
        self.lock = Lock()
        self.kwargs = kwargs
        self._no_update = False
        self.parameter_callback = parameter_callback
        self.escobj = escobj
        self.init_layout(**kwargs)
        self.on_displayed(self.update)
        
    def init_layout(self, **kwargs):
        self.center = self.init_center(**kwargs)
        self.header = self.init_header(**kwargs)
        self.footer = self.init_footer(**kwargs)
        self.left_sidebar = self.init_left_sidebar(**kwargs)
        self.right_sidebar = self.init_right_sidebar(**kwargs)
        
        if "pane_widths" not in kwargs: 
            self.pane_widths = [1, '650px', '300px']
        if "pane_heights" not in kwargs: 
            self.pane_heights = ['60px', '600px', '300px']
    
    def init_center(self, **kwargs):
        if self.center_widget_cls is not None:
            return self.center_widget_cls(self, **kwargs)
        return None
    
    def init_header(self, **kwargs):
        title = kwargs.get("title", '')
        if title:
            return widgets.HTML(
                value='<h2>{0}</h2>'.format(title))
        return None
    
    def init_footer(self, **kwargs):
        return None
    
    def init_left_sidebar(self, **kwargs):
        return None
    
    def init_right_sidebar(self, **kwargs):
        objpanel = ParametersPanel(
            self.escobj,
            layout=Layout(overflow_y='auto',
                          border='1px solid lightgrey',
                          padding="5px 1px 5px 10px"
                          ))
        for par in objpanel.children:
            par.slider.observe(self.parameter_callback, names='value')
        if not objpanel.children:
            return None
        return objpanel
    
    def update(self, *args):
        if self._no_update:
            return
        cw = self.center_widget
        if cw is not None:
            cw.update(*args)
    
    def update_parameter_sliders(self, silent=False):
        if silent:
            self._no_update = True 
        right_sidebar = self.right_sidebar
        if right_sidebar is not None:
            for par in right_sidebar.children:
                par.pull_value()
        if silent:
            self._no_update = False
           
    def disable_parameters(self, val):
        rsb = self.right_sidebar
        if rsb is not None:
            rsb.disabled = val
        
    

class PlotObjectLayout(EscapeObjectLayout):
    
    plot2d_widget_class = ErrorPlot2dWidget
    map_widget_class = MapWidget
    
    def init_center(self, **kwargs):
        escobj=self.escobj
        plot_type = kwargs.get(PLOT_TYPE_KW, PLOT2D_TYPE)
        if plot_type not in PLOT_TYPES:
            raise ValueError("Wrong type of plot, expected one of [{0}], got {1}".format(
                    ", ".join(PLOT_TYPES), plot_type
                ))
        if plot_type == MAP_TYPE:
            coordidx = kwargs.get("coord_index", [0, 1])
            if not isinstance(coordidx, (list, tuple)):
                raise TypeError("Unexpected type of coordinate index, expected list or tuple, got {0}".format(
                        type(coordidx)
                        ))
        else:
            coordidx = kwargs.get("coord_index", 0)
            if not isinstance(coordidx, int):
                raise TypeError("Unexpected type of coordinate index, expected int, got {0}".format(
                        type(coordidx)
                   ))
        rows = int(kwargs.get("rows", 1))
        if rows<1:
            raise ValueError("Number of rows cannot be <1")
        if plot_type == MAP_TYPE and rows==1:
            raise ValueError("Unexpected value for number of rows for 'map' plot type, got {0}".format(rows))
        self._plot_type = plot_type
        self._rows = rows
        self._coordidx = coordidx
        if plot_type==PLOT2D_TYPE:
            self.center_widget_cls = self.plot2d_widget_class
        elif plot_type==MAP_TYPE and escobj.domain_size>=2:
            self.center_widget_cls = self.map_widget_class
        else:
            raise ValueError("Cannot set proper plot class, check you settings")
        return super(PlotObjectLayout, self).init_center(**kwargs)
    