import numpy as np
import ipywidgets.widgets as widgets

from escape.utils.widgets.base import EscapeObjectLayout, PlotObjectLayout
from escape.utils.widgets.model import ModelObjectLayout
from escape.utils.widgets.plot import ModelPlot2dWidget, ModelMapWidget
from escape.utils.widgets.plot import PLOT_TYPE_KW, \
    PLOT2D_TYPE, MAP_TYPE, PLOT_TYPES
    
from escape.utils.widgets.base import lock_update
import warnings


class MSObjectLayout(EscapeObjectLayout):
    
    def init_layout(self, **kwargs):
        acc = widgets.Tab()
        num_mod = len(self.escobj)
        ch = []
        names = []
        kwlist = [{} for i in range(num_mod)]
        idx = 0
        for k, v in kwargs.items():
            if isinstance(v, (list, tuple)):
                if len(v) != num_mod:
                    msg = "Array length mismatch for argument {0}. Argument will be ignored.".format(k)
                    warnings.warn(msg)
                    continue
                for idx in range(num_mod):
                    kwlist[idx][k] = v[idx]
            else:
                for idx in range(num_mod):
                    kwlist[idx][k] = v
        for idx in range(num_mod):
            mdl = self.escobj.model(idx)
            names.append(mdl.name)
            mdlw = ModelObjectLayout(mdl, show_cost=False, **kwlist[idx])
            mdlw.header = None
            ch.append(mdlw)
        acc.children = ch
        for i, n in enumerate(names):
            acc.set_title(i, n)
        acc.observe(self.on_model_choose, names='selected_index')
        self.selected_index = 0
        self.center = acc
        self.header = self.init_header(**kwargs)
        self.footer = self.init_footer(**kwargs)
        
        if "pane_widths" not in kwargs: 
            self.pane_widths = [1, '650px', '300px']
        if "pane_heights" not in kwargs: 
            self.pane_heights = ['60px', '700px', '350px']
    
    @lock_update 
    def on_model_choose(self, *args):
        acc = args[0]['owner']
        selidx = acc.selected_index
        self.selected_index = selidx
        mdlw = self.center.children[selidx]
        mdlw.update_parameter_sliders(silent=True)
        mdlw._update(False)
     
    def update(self, *args):
        selidx = self.selected_index
        mdlw = self.center.children[selidx]
        mdlw.update(*args)
             
    def _update(self, buf=False):
        selidx = self.selected_index
        mdlw = self.center.children[selidx]
        mdlw._update(buf)
    
    def update_parameter_sliders(self, silent=False):
        selidx = self.selected_index
        mdlw = self.center.children[selidx]
        mdlw.update_parameter_sliders(silent)
        
    def disable_parameters(self, val):
        if self.center is not None:
            for c in self.center.children:
                c.disable_parameters(val)
        

class MSSinglePlotObjectLayout(PlotObjectLayout):
    
    plot2d_widget_class = ModelPlot2dWidget
    
    def _update(self, buf=False):
        escobj = self.escobj
        num_mod = len(escobj)
        self.center.reset_axes()
        for idx in range(num_mod):
            mdlobj = escobj.model(idx)
            self._update_model(mdlobj, buf)
        
    def _update_model(self, mdlobj, buf=False):
        dims = mdlobj.domain_size
        dobj = mdlobj.data
        xarr = np.array(dobj.ini_coordinates)
        yarr = np.array(dobj.ini_intensities)
        earr = np.array(dobj.ini_errors)
        if not len(earr):
            earr = np.sqrt(yarr)
        marr = np.array(dobj.mask)
        if not buf:
            mdlobj()
        sarr = np.array(mdlobj.simulation)
        rarr = np.array(mdlobj.residuals)
        
        if dims == 1:
            self.center.update(
                np.ma.masked_array(xarr, ~marr),
                np.ma.masked_array(yarr, ~marr),
                np.ma.masked_array(earr, ~marr),
                np.ma.masked_array(sarr, ~marr),
                np.ma.masked_array(rarr, ~marr),
                clearax=False,
                model_legend=mdlobj.name,
                data_legend=dobj.name)
        else:
            plot_type = self._plot_type
            coordidx = self._coordidx
            if plot_type == PLOT2D_TYPE:
                if not isinstance(coordidx, int):
                    raise TypeError("Unexpected type of coordinate index, expected int, got {0}".format(
                        type(coordidx)
                    ))
                xarr = xarr[coordidx::dims]
                self.center.update(
                    np.ma.masked_array(xarr, ~marr),
                    np.ma.masked_array(yarr, ~marr),
                    np.ma.masked_array(earr, ~marr),
                    np.ma.masked_array(sarr, ~marr),
                    np.ma.masked_array(rarr, ~marr),
                    clearax=False,
                    model_legend=mdlobj.name,
                    data_legend=dobj.name)
            else:
                raise ValueError("Wrong type of plot, expected one of [{0}], got {1}".format(
                    ", ".join(PLOT_TYPES), plot_type
                ))      
    
    def update(self, *args):
        if self._no_update:
            return
        self._update(False)
        
        
