import numpy as np
from matplotlib.figure import Figure
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.colors import LogNorm, Normalize

from ipympl.backend_nbagg import Canvas as Canvas, FigureManager
from matplotlib.gridspec import GridSpec
from matplotlib.transforms import Bbox
from typing import Tuple

PLOT_TYPE_KW = "plot_type"
PLOT2D_TYPE = "plot2d"
MAP_TYPE = "map"
PLOT_TYPES=[PLOT2D_TYPE, MAP_TYPE]


class BasePlotWidget(Canvas):
    
    __num__ = 0
    
    def __init__(self, parent, **kwargs):
        dpi = kwargs.get('dpi', 100)
        width = kwargs.get('width', 600)
        height = kwargs.get('height', 550)
        Canvas.__init__(self, Figure(dpi=dpi, figsize=(width/dpi, height/dpi)))
        self.resizable = False
        self.capture_scroll = True
        self.parent = parent
        self.axes = []
        self.init_axes(**kwargs)
        self.init_plots(**kwargs)
        self.apply_settings(**kwargs)
        self.kwargs=kwargs
        self.header_visible=False
        manager = FigureManager(self, self.__num__)
        def destroy(event):
            canvas.mpl_disconnect(cid)
            Gcf.destroy(manager)
        cid = self.mpl_connect('close_event', destroy)
        self.__num__+=1
        self._view_limits=[]
        
        
    def clear_axes(self):
        for ax in self.axes:
            ax.cla()
            
    def remove_axes(self):
        self._updated = False
        for ax in self.axes:
            ax.cla()
            ax.remove()
        self.axes = []
    
    def init_axes(self, **kwargs):
        pass
    
    def init_plots(self, **kwargs):
        pass
    
    def apply_settings(self, **kwargs):
        pass
    
    def reset_axes(self):
        self.clear_axes()
        self.apply_settings(**self.kwargs)
    
    def update(self, *args, **kwargs):
        pass
    

class Plot2dWidget(BasePlotWidget):
    
    def init_axes(self, **kwargs):
        if self.axes:
            self.remove_axes()
        fig = self.figure
        ax = fig.add_subplot(1, 1, 1)
        self.axes.append(ax)
        
    def apply_settings(self, **kwargs):
        ax = self.axes[0]
        if kwargs.get('grid', False):
            ax.grid()
        ax.set_xlabel(kwargs.get('xlabel', 'X'))
        ax.set_ylabel(kwargs.get('ylabel', 'Y'))
        if kwargs.get('xlog', False):
            ax.set_xscale('log')
        else:
            ax.set_xscale('linear')
        if kwargs.get('ylog', False):
            ax.set_yscale('log')
        else:
            ax.set_yscale('linear')
        figtitle = kwargs.get('figtitle', "")
        self.figure.suptitle(figtitle, fontsize=14)
        if kwargs.get('tight', False):
            self.figure.tight_layout()
    
    def init_plots(self, **kwargs):
        ax = self.axes[0]
        self.lines, = ax.plot([], [], 'o-', markersize=2, linewidth=2)
        
    def update(self, xarr, yarr):
        ax = self.axes[0]
        self.lines.set_xdata(xarr)
        self.lines.set_ydata(yarr)
        ax.relim()
        ax.autoscale_view()
        self.toolbar._update_view()
        self.draw_idle()
        
        
class ErrorPlot2dWidget(Plot2dWidget):
            
    def update(self, xarr, yarr, earr):
        self.clear_axes()
        self.apply_settings(**self.kwargs)
        ax = self.axes[0]
        self.errorfill(xarr, yarr, earr, None, 0.3, ax)
        self.draw_idle()
        
    def errorfill(self, x, y, yerr, color=None, alpha_fill=0.3, ax=None, label=""):
        """https://tonysyu.github.io/plotting-error-bars.html#.Xle85UoxlaQ"""
        ax = ax if ax is not None else self.figure.gca()
        if np.isscalar(yerr) or len(yerr) == len(y):
            ymin = y - yerr
            ymax = y + yerr
        elif len(yerr) == 2:
            ymin, ymax = yerr
        line, = ax.plot(x, y, color=color, marker='o',
                        fillstyle='none',
                        markersize=5, linestyle='None',
                        label=label)
        col = line.get_color()
        ax.fill_between(x, ymax, ymin, color=col, alpha=alpha_fill)
        return col
        


class MapWidget(BasePlotWidget):
    
    def init_axes(self, **kwargs):
        if self.axes:
            self.remove_axes()
        self.cbar = None
        fig = self.figure
        ax = fig.add_subplot(1, 1, 1)
        # create an axes on the right side of ax. The width of cax will be 5%
        # of ax and the padding between cax and ax will be fixed at 0.05 inch.
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        self.axes = [ax, cax]
    
    def update(self, zarr, extent=None):
        self.clear_axes()
        self.apply_settings(**self.kwargs)
        ax = self.axes[0]
        cax = self.axes[1]
        norm=None
        if self.cblog:
            norm=LogNorm(self.cbmin, self.cbmax)
        else:
            norm=Normalize(self.cbmin, self.cbmax)
        self.img = ax.imshow(zarr, origin='lower', extent=extent, aspect="auto",
                             norm=norm)
        self.cbar = self.figure.colorbar(self.img, cax=cax)
        self.draw_idle()
        
    def apply_settings(self, **kwargs):
        ax = self.axes[0]
        ax.set_xlabel(kwargs.get('xlabel', 'X'))
        ax.set_ylabel(kwargs.get('ylabel', 'Y'))
        self.cblog=kwargs.get('cblog', False)
        self.cbmin=kwargs.get('cbmin', None)
        self.cbmax=kwargs.get('cbmax', None)
        figtitle = kwargs.get('figtitle', "")
        self.figure.suptitle(figtitle, fontsize=14)
        if kwargs.get('tight', False):
            self.figure.tight_layout()
        
        
class ModelPlot2dWidget(ErrorPlot2dWidget):
    
    def init_axes(self, **kwargs):
        if self.axes:
            self.remove_axes()            
        fig = self.figure
        gs = GridSpec(2, 1, height_ratios=[4, 1])
        ax0 = fig.add_subplot(gs[0])
        ax1 = fig.add_subplot(gs[1], sharex=ax0)
        self.axes = [ax0, ax1]
        self.show_legend=kwargs.get("show_legend", True)
    
    def update(self, xarr, yarr, earr, sarr, rarr,
               clearax=True, model_legend="", data_legend=""):
        if clearax:
            self.clear_axes()
            self.apply_settings(**self.kwargs)
        ax = self.axes[0]
        self.errorfill(xarr, yarr, earr, None, 0.3, ax, label=data_legend)
        lines = ax.plot(xarr, sarr, lw=3, label=model_legend)
        diff = self.axes[1]
        diff.plot(xarr, rarr, color=lines[0].get_color(), lw=1)
        self.toolbar._update_view()
        ax.legend()
        self.draw_idle()
        
    def apply_settings(self, **kwargs):
        ax0 = self.axes[0]
        ax1 = self.axes[1]
        if kwargs.get('grid', False):
            ax0.grid()
        ax0.set_xlabel('')
        ax0.set_ylabel(kwargs.get('ylabel', 'Y'))
        ax1.set_xlabel(kwargs.get('xlabel', 'X'))
        ax1.set_ylabel('Residuals')
        if kwargs.get('xlog', False):
            ax0.set_xscale('log')
        else:
            ax0.set_xscale('linear')
        if kwargs.get('ylog', False):
            ax0.set_yscale('log')
        else:
            ax0.set_yscale('linear')
        figtitle = kwargs.get('figtitle', "")
        self.figure.suptitle(figtitle, fontsize=14)
        if kwargs.get('tight', False):
            self.figure.tight_layout()
            
    
        
class ModelMapWidget(MapWidget):
    
    def init_axes(self, **kwargs):
        if self.axes:
            self.remove_axes()
        self.cbar = None
        fig = self.figure
        gs = fig.add_gridspec(2, 2,  width_ratios=(7, 2), height_ratios=(2, 7),
                      left=0.1, right=0.9, bottom=0.1, top=0.9,
                      wspace=0.05, hspace=0.05)

        ax0 = fig.add_subplot(gs[1, 0])
        axx = fig.add_subplot(gs[0, 0], sharex=ax0)
        axy = fig.add_subplot(gs[1, 1], sharey=ax0)
        axx.tick_params(axis="x", labelbottom=False)
        axy.tick_params(axis="y", labelleft=False)
        self.axes = [ax0, axx, axy]
    
    def update(self, zarr_exp, zarr_sim, residarr, extent=None):
        self.clear_axes()
        if self.cbar is not None:
            self.cbar.remove()
        self.apply_settings(**self.kwargs)
        if self.cblog:
            norm=LogNorm(self.cbmin, self.cbmax)
        else:
            norm=Normalize(self.cbmin, self.cbmax)
        ax = self.axes[0]
        axx = self.axes[1]
        axy = self.axes[2]
        
        self.img0 = ax.imshow(zarr_exp, origin='lower', cmap="Greens", aspect="auto",
                              extent=extent, alpha=1.0, norm=norm)
        self.img1 = ax.imshow(zarr_sim, origin='lower', cmap="Oranges", aspect="auto",
                               extent=extent, alpha=0.5, norm=norm)
        residarr = np.abs(residarr)
        axx.plot(np.linspace(extent[0], extent[1], zarr_exp.shape[0]), np.sum(residarr, axis=0))
        axy.plot(np.sum(residarr, axis=1), np.linspace(extent[2], extent[3], zarr_exp.shape[1]))
        self.draw_idle()
        
    def apply_settings(self, **kwargs):
        ax = self.axes[0]
        ax.set_xlabel(kwargs.get('xlabel', 'X'))
        ax.set_ylabel(kwargs.get('ylabel', 'Y'))
        self.cblog=kwargs.get('cblog', False)
        self.cbmin=kwargs.get('cbmin', 1)
        self.cbmax=kwargs.get('cbmax', None)
        if kwargs.get('tight', False):
            self.figure.tight_layout()
        figtitle = kwargs.get('figtitle', "")
        self.figure.suptitle(figtitle, fontsize=14)
        
        
