import unittest
import numpy as np

from escape.core import *


class TestOptimizerMethods(unittest.TestCase):
    
    def test_lm(self):
        x = var("x")
        
        a0 = par("a0", 100, userlim=[50, 300])
        x0 = par("x0", 0.2, userlim=[0.1, 0.3])
        c0 = par("c0", 0.2, userlim=[0.1, 0.3])
        
        d0 = par("d0", 1.1, userlim=[0.55, 1.65])

        a1 = par("a1", 10, userlim=[5, 15])
        x1 = par("x1", 4, userlim=[2, 6])
        c1 = par("c1", 0.8, userlim=[0.4, 1.2])

        a2 = par("a2", 7, userlim=[5, 15])
        x2 = par("x2", 3, userlim=[2, 6])
        c2 = par("c2", 1.0, userlim=[0.4, 1.2])
        
        f1 = exp(-(x - x0) * (x - x0) / 2.0 / (c0 * c0)) * a0        \
            +exp(-(x - x1) * (x - x1) / 2.0 / (c1 * c1)) * a1 + d0

        f2 = exp(-(x - x0) * (x - x0) / 2.0 / (c0 * c0)) * a0       \
            +exp(-(x - x2) * (x - x2) / 2.0 / (c2 * c2)) * a2 + d0
        
        eval = kernel("evaluator", f1, True)
        eval2 = kernel("evaluator", f2, True)
 
        xarr = np.linspace(-10.0, 10.0, 500, dtype=float).copy()
        yarr = np.zeros(shape=xarr.shape).copy()
        yarr2 = np.zeros(shape=xarr.shape).copy()
         
        eval(xarr, yarr)
        eval2(xarr, yarr2)
          
        dataobj = data("test_data", xarr, yarr)
        dataobj2 = data("test_data", xarr, yarr2)
          
        modobj = model("model1", eval, dataobj, 'data', 'none')
        modobj2 = model("model2", eval2, dataobj2, 'data', 'none')
          
        stackobj = modelstack("stack", [modobj, modobj2])
          
        optimizer = levmar("LM", stackobj, 1e-10, 1e-10, 1e-10, 3000, 5000, 1)
          
        optimizer.shake()
          
        optimizer()
  
        self.assertAlmostEqual(a0.value, 100.0, delta=1e-5)
        self.assertAlmostEqual(x0.value, 0.2, delta=1e-5)
        self.assertAlmostEqual(c0.value, 0.2, delta=1e-5)
        self.assertAlmostEqual(d0.value, 1.1, delta=1e-5)
        self.assertAlmostEqual(a1.value, 10.0, delta=1e-5)
        self.assertAlmostEqual(x1.value, 4.0, delta=1e-5)
        self.assertAlmostEqual(c1.value, 0.8, delta=1e-5)
  
        self.assertAlmostEqual(a2.value, 7, delta=1e-5)
        self.assertAlmostEqual(x2.value, 3, delta=1e-5)
        self.assertAlmostEqual(c2.value, 1, delta=1e-5)   
    
    def test_de(self):
        x = var("x")
        
        a0 = par("a0", 100, userlim=[50, 300])
        x0 = par("x0", 0.2, userlim=[0.1, 0.3])
        c0 = par("c0", 0.2, userlim=[0.1, 0.3])
        
        d0 = par("d0", 1.1, userlim=[0.55, 1.65])

        a1 = par("a1", 10, userlim=[5, 15])
        x1 = par("x1", 4, userlim=[2, 6])
        c1 = par("c1", 0.8, userlim=[0.4, 1.2])

        a2 = par("a2", 7, userlim=[5, 15])
        x2 = par("x2", 3, userlim=[2, 6])
        c2 = par("c2", 1.0, userlim=[0.4, 1.2])
        
        f1 = exp(-(x - x0) * (x - x0) / 2.0 / (c0 * c0)) * a0        \
             +exp(-(x - x1) * (x - x1) / 2.0 / (c1 * c1)) * a1 + d0
 
        f2 = exp(-(x - x0) * (x - x0) / 2.0 / (c0 * c0)) * a0       \
             +exp(-(x - x2) * (x - x2) / 2.0 / (c2 * c2)) * a2 + d0
         
        eval = kernel("evaluator", f1, True)
        eval2 = kernel("evaluator", f2, True)
 
        xarr = np.linspace(-10.0, 10.0, 500, dtype=float).copy()
        yarr = np.zeros(shape=xarr.shape).copy()
        yarr2 = np.zeros(shape=xarr.shape).copy()
         
        eval(xarr, yarr)
        eval2(xarr, yarr2)
          
        dataobj = data("test_data", xarr, yarr)
        dataobj2 = data("test_data", xarr, yarr2)
          
        modobj = model("model1", eval, dataobj, 'data', 'none')
        modobj2 = model("model2", eval2, dataobj2, 'data', 'none')
          
        stackobj = modelstack("stack", [modobj, modobj2])
          
        optimizer = diffevol("DE", stackobj)
          
        optimizer.shake()
          
        optimizer()
  
        self.assertAlmostEqual(a0.value, 100.0, delta=1e-5)
        self.assertAlmostEqual(x0.value, 0.2, delta=1e-5)
        self.assertAlmostEqual(c0.value, 0.2, delta=1e-5)
        self.assertAlmostEqual(d0.value, 1.1, delta=1e-5)
        self.assertAlmostEqual(a1.value, 10.0, delta=1e-5)
        self.assertAlmostEqual(x1.value, 4.0, delta=1e-5)
        self.assertAlmostEqual(c1.value, 0.8, delta=1e-5)
  
        self.assertAlmostEqual(a2.value, 7, delta=1e-5)
        self.assertAlmostEqual(x2.value, 3, delta=1e-5)
        self.assertAlmostEqual(c2.value, 1, delta=1e-5)
         
        optimizer = diffevol("DE", stackobj, polish_candidate_maxiter=2, polish_final_maxiter=10)
          
        optimizer.shake()
          
        optimizer()
         
        self.assertAlmostEqual(a0.value, 100.0, delta=1e-5)
        self.assertAlmostEqual(x0.value, 0.2, delta=1e-5)
        self.assertAlmostEqual(c0.value, 0.2, delta=1e-5)
        self.assertAlmostEqual(d0.value, 1.1, delta=1e-5)
        self.assertAlmostEqual(a1.value, 10.0, delta=1e-5)
        self.assertAlmostEqual(x1.value, 4.0, delta=1e-5)
        self.assertAlmostEqual(c1.value, 0.8, delta=1e-5)
  
        self.assertAlmostEqual(a2.value, 7, delta=1e-5)
        self.assertAlmostEqual(x2.value, 3, delta=1e-5)
        self.assertAlmostEqual(c2.value, 1, delta=1e-5)


if __name__ == '__main__':
    unittest.main()
